////////////////////////////////////////////////////////////////////////////////
//                                  Include                                   //
////////////////////////////////////////////////////////////////////////////////


#include "Oracle.h"


////////////////////////////////////////////////////////////////////////////////
//                              Public: Oracle                                //
////////////////////////////////////////////////////////////////////////////////


Oracle::Oracle(char* filename) : x(env), u(env), aI(env), aJ(env), cx(env), gu(env),
				 obj(env), model(env), cplex(env)
{

  try {
    //----- User: Create data model -------------------------------------------1

    // Initiate variables
    IloInt i, j, aCount, aCol;


    // Data file
    ifstream file(filename);

    // m (rows), n (cols)
    file >> m >> n;

    cout << "m=" << m << "n=" << n <<"\n";
    
    // x (primal variables)
    x.add(IloNumVarArray(env, n, 0, 1));

    // u (dual multipliers)
    u.add(IloNumArray(env, m));
    // c (cost of building j)
    IloNumArray c(env, n);
//     for (j = 0; j < n; j++)
//       file >> c[j];

    IloNum sparsity;
    sparsity = 0.05;
    long  NN ;
    NN = (long)((double)m*(double)n*(double)sparsity);
     // NN = 9523545L;
    cout << "Total size of aI... " << NN << "\n";
    aJ.add(IloIntArray(env, n+1));
    aI.add(IloIntArray(env, NN));
    cout << "AI , AJ OK\n";
    long Jcur;
    Jcur = 0;
    // a (cover matrix)

    cout << "reading data file, this can take qxuite a while...  \n";

    for (j = 0; j < n; j++) {
      aJ[j] = Jcur;
      file >> c[j] >> aCount;
      for (i = 0; i < aCount; i++) {
 
        file >> aCol;

	aI[Jcur] = aCol -1;
	Jcur++;
      }
    }
    aJ[n] = Jcur;    
 
    // model: object
    for (j = 0; j < n; j++)
      cx += c[j] * x[j];

    // model: lagrange relaxad cover constraints
    IloExpr lag(env);

    // model: minimize
    obj = IloMinimize(env, cx + lag);
    model.add(obj);

    // model: x constraints
    for (j = 0; j < n; j++) {
      model.add(x[j] >= 0);
      model.add(x[j] <= 1);
    }

    // column fixing

    ifstream cf("ColumnFix");
    IloInt ncf;

    cf >> ncf;
    cout << "ncf="<<ncf << '\n';
    if (ncf > 0) {
      IloIntArray cfix(env, ncf);
      for (j = 0; j < ncf; j++)
	cf >> cfix[j];
      for (j = 0; j < ncf; j++){
	i = cfix[j]-1;
	model.remove(x[i] >= 0);
	model.remove(x[i] <= 1);
	model.add(   x[i] == 1 );
      }
    }
    
    // cplex: extract model
    cplex.extract(model);

    //-------------------------------------------------------------------------1
    //----- User: Variables that must be set ----------------------------------2

    // u size
    uSize = m;

    // x dimensions and size
    xDims = 1;
    xDim1Size = n;

    // Add all unconstrained multipliers u, end with an InINF
    ucVector = new unsigned int[1];
    ucVector[0] = InINF;

    // Is the primal problem of min type
    isMinProblem = true;

    //-------------------------------------------------------------------------2

    iteration = 0;

  } catch(IloException& e) {
    cerr << "Error: " << e.getMessage() << endl;
  } catch(...) {
    cout << "Error" << endl;
  }
}


////////////////////////////////////////////////////////////////////////////////
//                              Public: ~Oracle                               //
////////////////////////////////////////////////////////////////////////////////

Oracle::~Oracle()
{
  env.end();
}


////////////////////////////////////////////////////////////////////////////////
//                               Public: Solve                                //
////////////////////////////////////////////////////////////////////////////////


void Oracle::Solve(const double uVector[], double &fValue, double sgVector[],
		   double xVector[])
{
  //----- User: Solve model ----------------------------------------------------

  // Inititate variables
  IloInt i, j;
  IloNum sum;

  // Update u
  u.clear();
  for (i = 0; i < m; i++)
    u.add(uVector[i]);
 
  // Update model
  model.remove(obj);

  IloExpr lag(env);

//   for (i = 0; i < m; i++) {
//     IloExpr temp(env);
//     for (j = aI[i]; j < aI[i+1]; j++)
//       temp +=  x[aJ[j]];
//     lag += u[i] * (1 - temp);
//     temp.end();
//   }

  IloInt ii;
  double temp;
  for (j = 0; j < n; j++) {
    //    IloExpr temp(env);
    temp = 0.0;
    for (i = aJ[j]; i < aJ[j+1]; i++){
      ii = aI[i];
      temp +=  u[ii];
    }
    
    //     lag += u[i] * (1 - temp);
    lag += -1*x[j]*temp;
    // temp.end();
  }


  for (i = 0; i < m; i++)
    lag +=  u[i];

  obj = IloMinimize(env, cx + lag);
  lag.end();

  model.add(obj);

  // Solve model
  cplex.solve();

  // fValue (with "-" beacuse min problem)
  fValue = -cplex.getValue(obj);
  

  
  // sgVector (with "-" beacuse min problem)

//   for (i = 0; i < m; i++) {
//     sum = 0;
//     for (j = aI[i]; j < aI[i+1]; j++)
//       sum +=  cplex.getValue(x[aJ[j]]);
//     sgVector[i] = -(1 - sum);
//   }

  if (iteration==4) 
    cout <<"sg2 in the first run="<<sgVector[1] << endl;

  for (i = 0; i < m; i++)
    sgVector[i]=0.0;

  double xj;
  for (j = 0; j < n; j++) {
    for (i = aJ[j]; i < aJ[j+1]; i++){
      ii = aI[i];
      xj = cplex.getValue(x[j]);
      sgVector[ii] += xj;
    }
  
  }


  for (i = 0; i < m; i++)
      sgVector[i] -= 1;


  // xVector
  for (j = 0; j < n; j++)
    xVector[j] = cplex.getValue(x[j]);

  // Print iteration
  iteration += 1;
  cout << endl;
  cout << "Iteration " << iteration << " We want we want FORTRAN!!" << endl;
  
  if (iteration == 5) {
    cout << "f= "<< fValue << "+n";
    fValue = -cplex.getValue(cx);
    cout << "cx= "<< fValue << "+n";
    ofstream ufile("uVector5");
    for (i = 0; i < uSize; i++)
      ufile <<  i + 1 << " "<<  uVector[i] << endl;
    cout << "  u-vector written to file: uVector5" << endl;
    cout << endl;

    ofstream xfile("xVector5");
    for (j = 0; j < n; j++)
      xfile <<  j + 1 << " "<<  xVector[j] << endl;
    cout << "  x-vector written to file: xVector5" << endl;

    ofstream sgfile("sgVector5");

    for (i = 0; i < uSize; i++)
      sgfile <<  i + 1 << " "<<  sgVector[i] << endl;
    cout << "  sg-vector written to file: sgVector5" << endl;
    cout << endl;


  }
  
  /*
  cout << "------------------------------" << endl;
  cout << "  f = " << fValue << endl;
  for (i = 0; i < m; i++)
    cout << "  u" << i + 1 << " = " << u[i] << endl;
  for (j = 0; j < n; j++)
    cout << "  x" << j + 1 << " = " << cplex.getValue(x[j]) << endl;
  for (i = 0; i < m; i++)
    cout << "  g" << i + 1 << " = " << sgVector[i] << endl;
  cout << "------------------------------" << endl;
  cout << endl;
  */
  //----------------------------------------------------------------------------
}


////////////////////////////////////////////////////////////////////////////////
//                                Public: getX                                //
////////////////////////////////////////////////////////////////////////////////


long Oracle::getUSize()
{
  return uSize;
}

long Oracle::getXDims()
{
  return xDims;
}

long Oracle::getXDim1Size()
{
  return xDim1Size;
}

long Oracle::getXDim2Size()
{
  return xDim2Size;
}

bool Oracle::getIsMinProblem()
{
  return isMinProblem;
}

unsigned int* Oracle::getUCVector()
{
  return ucVector;
}


////////////////////////////////////////////////////////////////////////////////
